/*
 *                      ..::PictoGame::..
 *
 * Copyright (C) DKT 2010 <dktcoding [at] gmail>
 *
 * PictoGame is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or any later
 * version.
 *
 * PictoGame is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package picto.utils;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * This class represents the pattern file.<br>
 * The pattern files contain the following information:<br><ol>
 * <li><b>Title</b>  -- The name of the pattern
 * <li><b>Author</b> -- Who created the pattern
 * <li><b>Date</b>   -- The date the pattern was added to the game
 * <li><b>X size</b> -- Number of columns of the pattern
 * <li><b>y size</b> -- Number of rows of the pattern
 * <li><i>blank line</i>
 * <li><b>Data row 1</b>
 * <li><b>Data row 2</b>
 * <li><b>Data row ...</b>
 * <li><b>Data row n</b></ol>
 * The data is represented row by row with <tt>64-bit</tt> numbers in hex
 * (without the <tt>'0x'</tt> identifier), as you know each number is
 * represented by bits in this case each <tt>1</tt> means that the cell should
 * be painted and as you might gather each <tt>0</tt> means that the cell should
 * not be painted.
 *
 * @author Fido <dktcoding [at] gmail>
 */
public final class Picto implements Serializable {
    private static final long serialVersionUID = -535901729081793261L;
    private transient byte [][] columnHeaders;
    private transient byte [][] rowHeaders;
    private transient boolean[][] board;
    private transient long[] columns;
    private long[] data;
    private String title;
    private String author;
    private String date;
    private int numberOfRows;
    private int numberOfColumns;
    private int numberOfMarkedCells;

    public Picto(int i, int j){
        board = new boolean[i][j];
        numberOfRows = i;
        numberOfColumns = j;
    }

    /**
     * Creates a new Picto board from a file
     *
     * @param input input file
     * @param justInfo this flag tells the constructor if you are using
     * the Picto to get information or to play.
     */
    public Picto(InputStream input, boolean justInfo){
        loadBoard(new BufferedReader(new InputStreamReader(input)));
        if (!justInfo){
            loadHeaders();
        }
    }

    public Picto(BufferedReader buff){
        loadBoard(buff);
    }

    public void regenerate(){
        loadHeaders();
    }

    public void flush(){
        columnHeaders = null;
        rowHeaders = null;
        columns = null;
        board = null;
    }

    public boolean getElement(int i, int j){
        return board[i][j];
    }

    public long getRow(int i){
        return data[i];
    }

    public long getColumn(int i){
        if (columns == null){
            calculateColumns();
        }
        return columns[i];
    }

    public void setElement(int i, int j, boolean e){
        board[i][j] = e;
    }

    public int getColumnHeaderNum(){
        return columnHeaders[0].length;
    }

    public int getRowHeaderNum(){
        return rowHeaders[0].length;
    }

    public int getColumnHeadersElement(int i, int j) {
        return columnHeaders[i][j];
    }

    public int getRowHeadersElement(int i, int j) {
        return rowHeaders[i][j];
    }

    public int getNumberOfColumns() {
        return numberOfColumns;
    }

    public void setNumberOfColumns(int numberOfColumns) {
        this.numberOfColumns = numberOfColumns;
    }

    public int getNumberOfRows() {
        return numberOfRows;
    }

    public void setNumberOfRows(int numberOfRows) {
        this.numberOfRows = numberOfRows;
    }

    public String getTitle() {
        return title;
    }

    public void setTitle(String title) {
        this.title = title;
    }

    public String getAuthor() {
        return author;
    }

    public void setAuthor(String author) {
        this.author = author;
    }

    public String getDate() {
        return date;
    }

    public void setDate(String date) {
        this.date = date;
    }

    /**
     * Loads a pattern from a panel
     *
     * @param file the file to read
     */
    private void loadBoard(BufferedReader file){
        String line;
        try {
            setTitle (file.readLine().substring("# Title: " .length()));
            setAuthor(file.readLine().substring("# Author: ".length()));
            setDate  (file.readLine().substring("# Date: "  .length()));
            line = file.readLine().substring("# X size: ".length());
            setNumberOfColumns(Integer.valueOf(line));
            line = file.readLine().substring("# Y size: ".length());
            setNumberOfRows(Integer.valueOf(line));line = file.readLine();
            data = new long[numberOfRows];
            numberOfMarkedCells = 0;
            for(byte i = 0; i < numberOfRows; i++){
                data[i] = Long.valueOf(file.readLine(), 16);
                numberOfMarkedCells += Long.bitCount(data[i]);
            }
        } catch (IndexOutOfBoundsException ex) {
            Logger.getLogger(Picto.class.getName()).log(Level.SEVERE, null, ex);
        } catch (IOException ex) {
            Logger.getLogger(Picto.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    /**
     * Load the headers and the binary panel
     */
    private void loadHeaders(){
        if (data == null){
            loadData();
        }
        if (board == null){
            loadPanel();
        }
        loadColumnHeaders();
        loadRowHeaders();
    }

    /**
     * Loads the binary panel from the data
     */
    private void loadPanel(){
        board = new boolean[numberOfRows][numberOfColumns];
        for(int i = 0; i < numberOfRows; i++){
            for (int j = numberOfColumns-1, k = 0; j >= 0; j--){
                board[i][k++] = (data[i] & (1L << j)) != 0;
            }
        }
    }

    /**
     * Load the row headers from the data
     * @TODO clean up and optimize this a "little" bit
     */
    private void loadRowHeaders(){
        ArrayList<ArrayList<Byte>> rows = new ArrayList<ArrayList<Byte>>(numberOfRows);
        int max_e = 0;
        byte counter;
        for (boolean[] row : board){
            ArrayList<Byte> temp = new ArrayList<Byte>(5);
            counter = 0;
            for (boolean cell : row){
                if (cell){
                    counter++;
                } else if (counter != 0){
                    temp.add(counter);
                    counter = 0;
                }
            }
            if (counter != 0){
                temp.add(counter);
            }
            if (max_e < temp.size()){
                max_e = temp.size();
            }
            rows.add(temp);
        }
        counter = 0;
        rowHeaders = new byte[numberOfRows][max_e];
        for (ArrayList<Byte> foo : rows){
            for (int j = foo.size()-1, pos = max_e; j >= 0; j--){
                rowHeaders[counter][--pos] = foo.remove(j);
            }
            counter++;
        }

    }

    /**
     * Load the column headers from the data
     * @TODO clean up and optimize this a "little" bit
     */
    private void loadColumnHeaders(){
        ArrayList<ArrayList<Byte>> cols = new ArrayList
                                           <ArrayList<Byte>>(numberOfColumns);
        byte counter;
        int max_e = 0;
        for (int i = 0; i < numberOfColumns; i++){
            ArrayList<Byte> temp = new ArrayList<Byte>(5);
            counter = 0;
            for (int j = 0; j < numberOfRows; j++){
                if (board[j][i]){
                    counter++;
                } else if (counter != 0){
                    temp.add(counter);
                    counter = 0;
                }
            }
            if (counter != 0){
                temp.add(counter);
            }
            if (max_e < temp.size()){
                max_e = temp.size();
            }
            cols.add(temp);
        }
        counter = 0;
        columnHeaders = new byte[numberOfColumns][max_e];
        for (ArrayList<Byte> foo : cols){
            for (int j = foo.size()-1, pos = max_e; j >= 0; j--){
                columnHeaders[counter][--pos] = foo.remove(j);
            }
            counter++;
        }
    }

    /**
     * Returns the sum of all the marked cells
     *
     * @return number of marked cells
     */
    public int getNumberOfMarkedCells() {
        return numberOfMarkedCells;
    }

    /**
     * This method calculates the vertical data, is used by the solver
     */
    private void calculateColumns(){
        StringBuilder foo;
        if (board == null){
            regenerate();
        }
        columns = new long[numberOfRows];
        byte i, j;
        for (i = 0; i < numberOfColumns; i++){
            foo = new StringBuilder(numberOfRows);
            for (j = 0; j < numberOfRows; j++){
                foo.append(board[j][i] ? "1" : "0");
            }
            columns[i] = Long.parseLong(foo.toString(), 2);
        }
    }

    /**
     * Loads the data from the binary board, this method is called
     * when you save a newly created pattern
     */
    private void loadData(){
        StringBuilder foo;
        numberOfMarkedCells = 0;
        data = new long[numberOfRows];
        byte i, j;
        for (i = 0; i < numberOfRows; i++){
            foo = new StringBuilder(numberOfColumns);
            for (j = 0; j < numberOfColumns; j++){
                foo.append(board[i][j] ? "1" : "0");
            }
            data[i] = Long.parseLong(foo.toString(), 2);
            numberOfMarkedCells += Long.bitCount(data[i]);
        }
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        for (int i = 0; i < numberOfRows; i++){
            if (data[i] != ((Picto)obj).getRow(i)){
                return false;
            }
        }
        return true;
    }

    @Override
    public int hashCode() {
        return 41 * 7 + Arrays.hashCode(data);
    }

    /**
     * Generates the representation of a board to be saved as a playing board
     * @return board to save
     */
    public String toSavingString(){
        String str = "# Title: " + title +"\n"
                   + "# Author: " + author +"\n"
                   + "# Date: " + date +"\n"
                   + "# X size: " + numberOfColumns +"\n"
                   + "# Y size: " + numberOfRows +"\n\n";
        if (data == null){
            loadData();
        }
        for (int i = 0; i < numberOfRows; i++) {
            str += Long.toHexString(data[i]) + "\n";
        }
        return str;
    }

    @Override
    public String toString(){
        int size = numberOfRows * (numberOfColumns + 1);
        StringBuilder str = new StringBuilder(size);
        for (boolean[] row : board){
            for (boolean marked_cell : row){
                str.append(marked_cell ? "*" : "_");
            }
            str.append("\n");
        }
        return str.toString();
    }

}
